from .cobalt.cobaltLauncher import CobaltLauncher
from .local.local import LocalLauncher
from .lsf.lsfLauncher import LSFLauncher
from .pbs.pbsLauncher import PBSLauncher
from .slurm.slurmLauncher import SlurmLauncher

__all__ = [
    "CobaltLauncher",
    "LocalLauncher",
    "LSFLauncher",
    "PBSLauncher",
    "SlurmLauncher",
]



from .alpsSettings import AprunSettings
from .base import RunSettings
from .cobaltSettings import CobaltBatchSettings
from .lsfSettings import BsubBatchSettings, JsrunSettings
from .mpirunSettings import MpirunSettings
from .pbsSettings import QsubBatchSettings
from .slurmSettings import SbatchSettings, SrunSettings

__all__ = [
    "AprunSettings",
    "CobaltBatchSettings",
    "BsubBatchSettings",
    "JsrunSettings",
    "MpirunSettings",
    "QsubBatchSettings",
    "RunSettings",
    "SbatchSettings",
    "SrunSettings",
]



# BSD 2-Clause License
#
# Copyright (c) 2021-2022, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

# Constants for SmartSim

from warnings import warn, simplefilter


dep_msg = "This is a deprecated module. Please use smartsim.status instead.\n"
dep_msg += "This module will be removed in the next release."

simplefilter("once", DeprecationWarning)
warn(dep_msg, DeprecationWarning)


# Statuses that are applied to jobs
STATUS_RUNNING = "Running"
STATUS_COMPLETED = "Completed"
STATUS_CANCELLED = "Cancelled"
STATUS_FAILED = "Failed"
STATUS_NEW = "New"
STATUS_PAUSED = "Paused"

# SmartSim status mapping
SMARTSIM_STATUS = {
    "Running": STATUS_RUNNING,
    "Paused": STATUS_PAUSED,
    "Completed": STATUS_COMPLETED,
    "Cancelled": STATUS_CANCELLED,
    "Failed": STATUS_FAILED,
    "New": STATUS_NEW,
}



import os
import re
import shutil
import stat
import subprocess
import sys
from pathlib import Path
from subprocess import SubprocessError

# NOTE: This will be imported by setup.py and hence no
#       smartsim related items should be imported into
#       this file.

# TODO:
#   - check cmake version and use system if possible to avoid conflicts


class BuildError(Exception):
    pass


class Builder:
    """Base class for building third-party libraries"""

    url_regex = re.compile(
        r"^(?:http|ftp)s?://"  # http:// or https://
        r"(?:(?:[A-Z0-9](?:[A-Z0-9-]{0,61}[A-Z0-9])?\.)+(?:[A-Z]{2,6}\.?|[A-Z0-9-]{2,}\.?)|"  # domain...
        r"localhost|"  # localhost...
        r"\d{1,3}\.\d{1,3}\.\d{1,3}\.\d{1,3})"  # ...or ip
        r"(?::\d+)?"  # optional port
        r"(?:/?|[/?]\S+)$",
        re.IGNORECASE,
    )

    def __init__(self, env, jobs=1, verbose=False):

        # build environment from buildenv
        self.env = env

        # Find _core directory and set up paths
        _core_dir = Path(os.path.abspath(__file__)).parent.parent
        self.build_dir = _core_dir / ".third-party"
        self.bin_path = _core_dir / "bin"
        self.lib_path = _core_dir / "lib"

        # Set wether build process will output to std output
        self.out = subprocess.DEVNULL
        self.verbose = verbose
        if self.verbose:
            self.out = None

        # make build directory "SmartSim/smartsim/_core/.third-party"
        if not self.build_dir.is_dir():
            self.build_dir.mkdir()
        if not self.bin_path.is_dir():
            self.bin_path.mkdir()
        if not self.lib_path.is_dir():
            self.lib_path.mkdir()

        self.jobs = jobs

    # implemented in base classes
    @property
    def is_built(self):
        raise NotImplementedError

    def build_from_git(self):
        raise NotImplementedError

    def binary_path(self, binary):
        binary_ = shutil.which(binary)
        if binary_:
            return binary_
        raise BuildError(f"{binary} not found in PATH")

    def copy_file(self, src, dst, set_exe=False):
        shutil.copyfile(src, dst)
        if set_exe:
            Path(dst).chmod(stat.S_IXUSR | stat.S_IWUSR | stat.S_IRUSR)

    def copy_dir(self, src, dst, set_exe=False):
        src = Path(src)
        dst = Path(dst)
        dst.mkdir(exist_ok=True)
        # copy directory contents
        for content in src.glob("*"):
            if content.is_dir():
                self.copy_dir(content, dst / content.name, set_exe=set_exe)
            else:
                self.copy_file(content, dst / content.name, set_exe=set_exe)

    def is_valid_url(self, url):
        return re.match(self.url_regex, url) is not None

    def cleanup(self):
        if self.build_dir.is_dir():
            shutil.rmtree(str(self.build_dir))

    def run_command(self, cmd, shell=False, out=None, cwd=None):
        # option to manually disable output if necessary
        if not out:
            out = self.out
        try:
            proc = subprocess.Popen(
                cmd,
                stderr=subprocess.PIPE,
                stdout=out,
                cwd=cwd,
                shell=shell,
                env=self.env,
            )
            error = proc.communicate()[1].decode("utf-8")
            if proc.returncode != 0:
                raise BuildError(error)
        except (OSError, SubprocessError) as e:
            raise BuildError(e)


class RedisBuilder(Builder):
    """Class to build Redis from Source
    Supported build methods:
     - from git
    See buildenv.py for buildtime configuration of Redis
    version and url.
    """

    def __init__(self, build_env={}, malloc="libc", jobs=None, verbose=False):
        super().__init__(build_env, jobs=jobs, verbose=verbose)
        self.malloc = malloc

    @property
    def is_built(self):
        """Check if Redis is built"""
        server = self.bin_path.joinpath("redis-server").is_file()
        cli = self.bin_path.joinpath("redis-cli").is_file()
        return server and cli

    def build_from_git(self, git_url, branch):
        """Build Redis from git
        :param git_url: url from which to retrieve Redis
        :type git_url: str
        :param branch: branch to checkout
        :type branch: str
        """
        redis_build_path = Path(self.build_dir, "redis")

        # remove git directory if it exists as it should
        # really never exist as we delete after build
        if redis_build_path.is_dir():
            shutil.rmtree(str(redis_build_path))

        # Check Redis URL
        if not self.is_valid_url(git_url):
            raise BuildError(f"Malformed Redis URL: {git_url}")

        # clone Redis
        clone_cmd = [
            self.binary_path("git"),
            "clone",
            git_url,
            "--branch",
            branch,
            "--depth",
            "1",
            "redis",
        ]
        self.run_command(clone_cmd, cwd=self.build_dir)

        # build Redis
        build_cmd = [
            self.binary_path("make"),
            "-j",
            str(self.jobs),
            f"MALLOC={self.malloc}",
        ]
        self.run_command(build_cmd, cwd=str(redis_build_path))

        # move redis binaries to smartsim/smartsim/_core/bin
        redis_src_dir = redis_build_path / "src"
        self.copy_file(
            redis_src_dir / "redis-server", self.bin_path / "redis-server", set_exe=True
        )
        self.copy_file(
            redis_src_dir / "redis-cli", self.bin_path / "redis-cli", set_exe=True
        )


class RedisAIBuilder(Builder):
    """Class to build RedisAI from Source
    Supported build method:
     - from git
    See buildenv.py for buildtime configuration of RedisAI
    version and url.
    """

    def __init__(
        self,
        build_env={},
        torch_dir="",
        build_torch=True,
        build_tf=True,
        build_onnx=False,
        jobs=None,
        verbose=False,
    ):
        super().__init__(build_env, jobs=jobs, verbose=verbose)
        self.rai_build_path = Path(self.build_dir, "RedisAI")

        # convert to int for RAI build script
        self.torch = 1 if build_torch else 0
        self.tf = 1 if build_tf else 0
        self.onnx = 1 if build_onnx else 0
        self.torch_dir = torch_dir

    @property
    def is_built(self):
        server = self.lib_path.joinpath("backends").is_dir()
        cli = self.lib_path.joinpath("redisai.so").is_file()
        return server and cli

    def copy_tf_cmake(self):
        """Copy the FindTensorFlow.cmake file to the build directory
        as the version included in RedisAI is out of date for us.
        Note: opt/cmake/modules removed in RedisAI v1.2.5
        """
        # remove the previous version
        tf_cmake = self.rai_build_path / "opt/cmake/modules/FindTensorFlow.cmake"
        tf_cmake.resolve()
        if tf_cmake.is_file():
            tf_cmake.unlink()
            # copy ours in
            self.copy_file(
                self.bin_path / "modules/FindTensorFlow.cmake", tf_cmake, set_exe=False
            )

    def build_from_git(self, git_url, branch, device):
        """Build RedisAI from git
        :param git_url: url from which to retrieve RedisAI
        :type git_url: str
        :param branch: branch to checkout
        :type branch: str
        :param device: cpu or gpu
        :type device: str
        """

        # delete previous build dir (should never be there)
        if self.rai_build_path.is_dir():
            shutil.rmtree(self.rai_build_path)

        # Check RedisAI URL
        if not self.is_valid_url(git_url):
            raise BuildError(f"Malformed RedisAI URL: {git_url}")

        # clone RedisAI
        clone_cmd = [
            self.binary_path("env"),
            "GIT_LFS_SKIP_SMUDGE=1",
            "git",
            "clone",
            "--recursive",
            git_url,
            "--branch",
            branch,
            "--depth=1",
            "RedisAI",
        ]
        self.run_command(clone_cmd, out=subprocess.DEVNULL, cwd=self.build_dir)

        # copy FindTensorFlow.cmake to RAI cmake dir
        self.copy_tf_cmake()

        # get RedisAI dependencies
        dep_cmd = [
            self.binary_path("env"),
            f"WITH_PT=0",  # torch is always 0 because we never use the torch from RAI
            f"WITH_TF={self.tf}",
            f"WITH_TFLITE=0",  # never build with TF lite (for now)
            f"WITH_ORT={self.onnx}",
            "VERBOSE=1",
            self.binary_path("bash"),
            self.rai_build_path / "get_deps.sh",
            device,
        ]

        self.run_command(
            dep_cmd,
            out=subprocess.DEVNULL,  # suppress this as it's not useful
            cwd=self.rai_build_path,
        )

        build_cmd = [
            self.binary_path("env"),
            f"WITH_PT={self.torch}",  # but we built it in if the user specified it
            f"WITH_TF={self.tf}",
            f"WITH_TFLITE=0",  # never build TF Lite
            f"WITH_ORT={self.onnx}",
            "WITH_UNIT_TESTS=0",
        ]

        if device == "gpu":
            build_cmd.append("GPU=1")
        else:
            build_cmd.append("GPU=0")

        if self.torch_dir:
            self.env["Torch_DIR"] = str(self.torch_dir)

        build_cmd.extend(
            [
                self.binary_path("make"),
                "-C",
                str(self.rai_build_path / "opt"),
                "-j",
                f"{self.jobs}",
                "build",
            ]
        )
        self.run_command(build_cmd, cwd=self.rai_build_path)

        self._install_backends(device)
        if self.torch:
            self._move_torch_libs()
        self.cleanup()

    def _install_backends(self, device):
        """Move backend libraries to smartsim/_core/lib/
        :param device: cpu or cpu
        :type device: str
        """
        self.rai_install_path = self.rai_build_path.joinpath(
            f"install-{device}"
        ).resolve()
        rai_lib = self.rai_install_path / "redisai.so"
        rai_backends = self.rai_install_path / "backends"

        if rai_lib.is_file() and rai_backends.is_dir():
            self.copy_dir(rai_backends, self.lib_path / "backends", set_exe=True)
            self.copy_file(rai_lib, self.lib_path / "redisai.so", set_exe=True)

    def _move_torch_libs(self):
        """Move pip install torch libraries
        Since we use pip installed torch libraries for building
        RedisAI, we need to move them into the LD_runpath of redisai.so
        in the smartsim/_core/lib directory.
        """

        ss_rai_torch_path = self.lib_path / "backends" / "redisai_torch"
        ss_rai_torch_lib_path = ss_rai_torch_path / "lib"

        # retrieve torch shared libraries and copy to the
        # smartsim/_core/lib/backends/redisai_torch/lib dir
        # self.torch_dir should be /path/to/torch/share/cmake/Torch
        # so we take the great grandparent here
        pip_torch_path = Path(self.torch_dir).parent.parent.parent
        pip_torch_lib_path = pip_torch_path / "lib"

        self.copy_dir(pip_torch_lib_path, ss_rai_torch_lib_path, set_exe=True)

        # also move the openmp files if on a mac
        if sys.platform == "darwin":
            dylibs = pip_torch_path / ".dylibs"
            self.copy_dir(dylibs, ss_rai_torch_path / ".dylibs", set_exe=True)



import numpy as np
import torch

from smartsim.ml.data import DynamicDataDownloader, StaticDataDownloader


class StaticDataGenerator(StaticDataDownloader, torch.utils.data.IterableDataset):
    """A class to download a dataset from the DB.

    Details about parameters and features of this class can be found
    in the documentation of ``StaticDataDownloader``, of which it is just
    a PyTorch-specialized sub-class.

    Note that if the ``StaticDataGenerator`` has to be used through a ``DataLoader``,
    `init_samples` must be set to `False`, as sources and samples will be initialized
    by the ``DataLoader`` workers.
    """

    def __init__(self, **kwargs):
        StaticDataDownloader.__init__(self, **kwargs)

    def _add_samples(self, batch_name, target_name):
        if self.samples is None:
            self.samples = torch.tensor(self.client.get_tensor(batch_name))
            if self.need_targets:
                self.targets = torch.tensor(self.client.get_tensor(target_name))
        else:
            self.samples = torch.cat(
                (self.samples, torch.tensor(self.client.get_tensor(batch_name)))
            )
            if self.need_targets:
                self.targets = torch.cat(
                    (self.targets, torch.tensor(self.client.get_tensor(target_name)))
                )

        self.num_samples = self.samples.shape[0]
        self.indices = np.arange(self.num_samples)
        self.log("Success!")
        self.log(f"New dataset size: {self.num_samples}")

    def update_data(self):
        self._update_samples_and_targets()
        if self.shuffle:
            np.random.shuffle(self.indices)


class DynamicDataGenerator(DynamicDataDownloader, StaticDataGenerator):
    """A class to download batches from the DB.

    Details about parameters and features of this class can be found
    in the documentation of ``DynamicDataDownloader``, of which it is just
    a PyTorch-specialized sub-class.

    Note that if the ``DynamicDataGenerator`` has to be used through a ``DataLoader``,
    `init_samples` must be set to `False`, as sources and samples will be initialized
    by the ``DataLoader`` workers.
    """

    def __init__(self, **kwargs):
        StaticDataGenerator.__init__(self, **kwargs)

    def __iter__(self):
        if self.sources:
            self.update_data()
        return super().__iter__()

    def _add_samples(self, batch_name, target_name):
        StaticDataGenerator._add_samples(self, batch_name, target_name)

    def __iter__(self):
        if self.sources:
            self.update_data()
        return super().__iter__()


class DataLoader(torch.utils.data.DataLoader):  # pragma: no cover
    """DataLoader to be used as a wrapper of StaticDataGenerator or DynamicDataGenerator

    This is just a sub-class of ``torch.utils.data.DataLoader`` which
    sets up sources of a data generator correctly. DataLoader parameters
    such as `num_workers` can be passed at initialization. `batch_size` should always
    be set to None.
    """

    def __init__(self, dataset: StaticDataGenerator, **kwargs):
        super().__init__(
            dataset,
            worker_init_fn=self.worker_init_fn,
            persistent_workers=True,
            **kwargs,
        )

    @staticmethod
    def worker_init_fn(worker_id):
        worker_info = torch.utils.data.get_worker_info()
        dataset = worker_info.dataset  # the dataset copy in this worker process
        dataset.init_sources()
        overall_sources = dataset.sources

        worker_id = worker_info.id

        # configure the dataset to only process the split workload
        per_worker = int((len(overall_sources)) // worker_info.num_workers)

        if per_worker > 0:
            if worker_id < worker_info.num_workers - 1:
                sources = overall_sources[
                    worker_id * per_worker : (worker_id + 1) * per_worker
                ]
            else:
                sources = overall_sources[worker_id * per_worker :]
        else:
            if worker_id < len(overall_sources):
                sources = overall_sources[worker_id]
            else:
                sources = []

        dataset.init_samples(sources)



from shutil import which
from subprocess import run


def detect_launcher():
    """Detect available launcher."""
    # Precedence: PBS, Cobalt, LSF, Slurm, local
    if which("qsub") and which("qstat") and which("qdel"):
        qsub_version = run(
            ["qsub", "--version"], shell=False, capture_output=True, encoding="utf-8"
        )
        if "pbs" in (qsub_version.stdout).lower():
            return "pbs"
        if "cobalt" in (qsub_version.stdout).lower():
            return "cobalt"
    if (
        which("bsub")
        and which("jsrun")
        and which("jslist")
        and which("bjobs")
        and which("bkill")
    ):
        return "lsf"
    if (
        which("sacct")
        and which("srun")
        and which("salloc")
        and which("sbatch")
        and which("scancel")
        and which("sstat")
        and which("sinfo")
    ):
        return "slurm"
    return "local"



from .errors import (
    AllocationError,
    EntityExistsError,
    LauncherError,
    ParameterWriterError,
    ShellError,
    SmartSimError,
    SSConfigError,
    SSInternalError,
    SSUnsupportedError,
    UserStrategyError,
)



import argparse
import shutil
import sys

from smartsim._core._cli.utils import get_install_path
from smartsim.log import get_logger

smart_logger_format = "[%(name)s] %(levelname)s %(message)s"
logger = get_logger("Smart", fmt=smart_logger_format)


class Clean:
    def __init__(self, clean_all=False):
        parser = argparse.ArgumentParser(
            description="Remove previous ML runtime installation"
        )
        parser.add_argument(
            "--clobber",
            action="store_true",
            default=False,
            help="Remove all SmartSim non-python dependencies as well",
        )
        args = parser.parse_args(sys.argv[2:])

        self._core_path = get_install_path() / "_core"
        clobber = args.clobber or clean_all
        self.clean(_all=clobber)

    def clean(self, _all=False):
        """Remove pre existing installations of ML runtimes

        :param _all: Remove all non-python dependencies
        :type _all: bool, optional
        """

        build_temp = self._core_path / ".third-party"
        if build_temp.is_dir():
            shutil.rmtree(build_temp, ignore_errors=True)

        lib_path = self._core_path / "lib"
        if lib_path.is_dir():

            # remove RedisAI
            rai_path = lib_path / "redisai.so"
            if rai_path.is_file():
                rai_path.unlink()
                logger.info("Successfully removed existing RedisAI installation")

            backend_path = lib_path / "backends"
            if backend_path.is_dir():
                shutil.rmtree(backend_path, ignore_errors=True)
                logger.info("Successfully removed ML runtimes")

        bin_path = self._core_path / "bin"
        if bin_path.is_dir() and _all:
            files_to_remove = ["redis-server", "redis-cli"]
            removed = False
            for _file in files_to_remove:
                file_path = bin_path.joinpath(_file)

                if file_path.is_file():
                    removed = True
                    file_path.unlink()
            if removed:
                logger.info("Successfully removed SmartSim Redis installation")



import time
from os import environ

import numpy as np
from smartredis import Client, Dataset

from ..error import SmartSimError
from ..log import get_logger

logger = get_logger(__name__)


def form_name(*args):
    return "_".join(str(arg) for arg in args if arg is not None)


class TrainingDataUploader:
    """A class to simplify uploading batches of samples to train a model.

    This class can be used to upload samples following a simple convention
    for naming. Once created, the function `publish_info` can be used
    to put all details about the data set on the Orchestrator. A training
    process can thus access them and get all relevant information to download
    the batches which are uploaded.

    Each time a new batch is available, it is sufficient to call `put_batch`,
    and the data will be stored following the naming convention specified
    by the attributes of this class.

    :param name: Name of the dataset as stored on the Orchestrator
    :type name: str
    :param sample_prefix: Prefix of samples batches
    :type sample_prefix: str
    :param target_prefix: Prefix of target batches (if needed)
    :type target_prefix: str
    :param num_classes: Number of classes of targets, if categorical
    :type num_classes: int
    :param producer_prefixes: Prefixes of processes which will be producing batches.
                              This can be useful in case the consumer processes also
                              have other incoming entities.
    :type producer_prefixes: str or list[str]
    :param cluster: Whether the SmartSim Orchestrator is being run as a cluster
    :type cluster: bool
    :param address: Address of Redis DB as <ip_address>:<port>
    :type address: str
    :param num_ranks: Number of processes (e.g. MPI ranks) of application using
                      DataUploader.
    :type num_ranks: int
    :param rank: Rank of DataUploader in multi-process application (e.g. MPI rank).
    :type rank: int
    :param verbose: If output should be logged to screen.
    :type verbose: bool

    """

    def __init__(
        self,
        name="training_data",
        sample_prefix="samples",
        target_prefix="targets",
        num_classes=None,
        producer_prefixes=None,
        cluster=True,
        address=None,
        num_ranks=None,
        rank=None,
        verbose=False,
    ):
        if not name:
            raise ValueError("Name can not be empty")
        if not sample_prefix:
            raise ValueError("Sample prefix can not be empty")

        self.name = name
        self.sample_prefix = sample_prefix
        self.target_prefix = target_prefix
        if isinstance(producer_prefixes, str):
            producer_prefixes = [producer_prefixes]
        self.producer_prefixes = producer_prefixes
        self.num_classes = num_classes
        if num_ranks is None:
            self.num_ranks = None
        else:
            self.num_ranks = int(num_ranks)
        self.rank = rank

        self.client = Client(address=address, cluster=cluster)
        self.batch_idx = 0
        self.verbose = verbose

    def publish_info(self):
        info_ds = Dataset(form_name(self.name, "info"))
        info_ds.add_meta_string("sample_prefix", self.sample_prefix)
        if self.target_prefix:
            info_ds.add_meta_string("target_prefix", self.target_prefix)
        if self.producer_prefixes:
            for producer_prefix in self.producer_prefixes:
                info_ds.add_meta_string("producer_prefixes", producer_prefix)
        if self.num_classes:
            info_ds.add_meta_scalar("num_classes", self.num_classes)
        if self.num_ranks:
            info_ds.add_meta_scalar("num_ranks", self.num_ranks)
        self.client.put_dataset(info_ds)

    def put_batch(self, samples, targets=None):
        batch_key = form_name(self.sample_prefix, self.batch_idx, self.rank)
        self.client.put_tensor(batch_key, samples)
        if self.verbose:
            logger.info(f"Put batch {batch_key}")

        if (
            targets is not None
            and self.target_prefix
            and (self.target_prefix != self.sample_prefix)
        ):
            labels_key = form_name(self.target_prefix, self.batch_idx, self.rank)
            self.client.put_tensor(labels_key, targets)

        self.batch_idx += 1


class StaticDataDownloader:
    """A class to download a dataset from the DB.

    By default, the StaticDataDownloader has to be created in a process
    launched through SmartSim, with sample producers listed as incoming
    entities.

    All details about the batches must be defined in
    the constructor; two mechanisms are available, `manual` and
    `auto`.

     - When specifying `auto`, the user must also specify
      `uploader_name`. StaticDataDownloader will get all needed information
      from the database (this expects a Dataset like the one created
      by TrainingDataUploader to be available and stored as `uploader_name`
      on the DB).

     - When specifying `manual`, the user must also specify details
       of batch naming. Specifically, for each incoming entity with
       a name starting with an element of `producer_prefixes`,
       StaticDataDownloader will query the DB
       for all batches named <sample_prefix>_<sub_index> for all indices
       in `sub_indexes` if supplied, and, if
       `target_prefix` is supplied, it will also query for all targets
       named <target_prefix>.<sub_index>. If `producer_prefixes` is
       None, then all incoming entities will be treated as producers,
       and for each one, the corresponding batches will be downloaded.

    The flag `init_samples` defines whether sources (the list of batches
    to be fetched) and samples (the actual data) should automatically
    be set up in the costructor.

    If the user needs to modify the list of sources, then `init_samples=False`
    has to be set. In that case, to set up a `BatchDownlaoder`, the user has to call
    `init_sources()` (which initializes the list of sources and the SmartRedis client)
    and `init_samples()`.  After `init_sources()` is called,
    a list of data sources is populated, representing the batches which
    will be downloaded.

    Each source is represented as a tuple `(producer_name, sub_index)`.
    Before `init_samples()` is called, the user can modify the list.
    Once `init_samples()` is called, all data is downloaded and batches
    can be obtained with iter().

    After initialization, samples and targets will not be updated. The data can
    be shuffled by calling `update_data()`, if `shuffle` is set to ``True`` at
    initialization.

    :param batch_size: Size of batches obtained with __iter__
    :type batch_size: int
    :param shuffle: whether order of samples has to be shuffled when calling `update_data`
    :type shuffle: bool
    :param uploader_info: Set to `auto` uploader information has to be downloaded from DB,
                          or to `manual` if it is provided by the user
    :type uploader_info: str
    :param uploader_name: Name of uploader info dataset, only used if `uploader_info` is `auto`
    :type uploader_name: str
    :param sample_prefix: prefix of keys representing batches
    :type sample_prefix: str
    :param target_prefix: prefix of keys representing targets
    :type target_prefix: str
    :param uploader_ranks: Number of processes every uploader runs on (e.g, if each
                           rank in an MPI simulation is uploading its own batches,
                           this will be the MPI comm world size of the simulation).
    :type uploader_ranks: int
    :param num_classes: Number of classes of targets, if categorical
    :type num_classes: int
    :param producer_prefixes: Prefixes of names of which will be producing batches.
                              These can be e.g. prefixes of SmartSim entity names in
                              an ensemble.
    :type producer_prefixes: str
    :param cluster: Whether the Orchestrator will be run as a cluster
    :type cluster: bool
    :param address: Address of Redis client as <ip_address>:<port>
    :type address: str
    :param replica_rank: When StaticDataDownloader is used distributedly, indicates
                         the rank of this object
    :type replica_rank: int
    :param num_replicas: When BatchDownlaoder is used distributedly, indicates
                         the total number of ranks
    :type num_replicas: int
    :param verbose: Whether log messages should be printed
    :type verbose: bool
    :param init_samples: whether samples should be initialized in the constructor
    :type init_samples: bool
    """

    def __init__(
        self,
        batch_size=32,
        shuffle=True,
        uploader_info="auto",
        uploader_name="training_data",
        sample_prefix="samples",
        target_prefix="targets",
        uploader_ranks=None,
        num_classes=None,
        producer_prefixes=None,
        cluster=True,
        address=None,
        replica_rank=0,
        num_replicas=1,
        verbose=False,
        init_samples=True,
        **kwargs,
    ):
        self.replica_rank = replica_rank
        self.num_replicas = num_replicas
        self.address = address
        self.cluster = cluster
        self.uploader_info = uploader_info
        self.uploader_name = uploader_name
        self.verbose = verbose
        self.samples = None
        self.targets = None
        self.num_samples = 0
        self.indices = np.arange(0)
        self.shuffle = shuffle
        self.batch_size = batch_size
        if uploader_info == "manual":
            self.sample_prefix = sample_prefix
            self.target_prefix = target_prefix
            if uploader_ranks is not None:
                self.sub_indices = [str(rank) for rank in range(uploader_ranks)]
            else:
                self.sub_indices = None
            if producer_prefixes:
                self.producer_prefixes = list(producer_prefixes)
            else:
                producer_prefixes = [""]
            self.num_classes = num_classes
        elif self.uploader_info == "auto":
            pass
        else:
            raise ValueError(
                f"uploader_info must be one of 'auto' or 'manual', but was {self.uploader_info}"
            )

        if init_samples:
            self.init_sources()
            self.init_samples()
        else:
            self.client = Client(self.address, self.cluster)
            if self.uploader_info == "auto":
                if not self.uploader_name:
                    raise ValueError(
                        "uploader_name can not be empty if uploader_info is 'auto'"
                    )
                self._get_uploader_info(self.uploader_name)
            # This avoids problems with Pytorch
            self.client = None

    def log(self, message):
        if self.verbose:
            logger.info(message)

    def _list_all_sources(self):
        uploaders = environ["SSKEYIN"].split(",")
        sources = []
        for uploader in uploaders:
            if any(
                [
                    uploader.startswith(producer_prefix)
                    for producer_prefix in self.producer_prefixes
                ]
            ):
                if self.sub_indices:
                    sources.extend(
                        [[uploader, sub_index] for sub_index in self.sub_indices]
                    )
                else:
                    sources.append([uploader, None])

        per_replica = len(sources) // self.num_replicas
        if per_replica > 0:
            if self.replica_rank < self.num_replicas - 1:
                sources = sources[
                    self.replica_rank
                    * per_replica : (self.replica_rank + 1)
                    * per_replica
                ]
            else:
                sources = sources[self.replica_rank * per_replica :]
        else:
            self.log(
                "Number of loader replicas is higher than number of sources, automatic split cannot be performed, "
                "all replicas will have the same dataset. If this is not intended, then implement a distribution strategy "
                "and modify `sources`."
            )

        return sources

    def init_sources(self):
        """Initalize list of data sources based on incoming entitites and self.sub_indices.


        Each source is represented as a tuple `(producer_name, sub_index)`.
        Before `init_samples()` is called, the user can modify the list.
        Once `init_samples()` is called, all data is downloaded and batches
        can be obtained with iter(). The list of all sources is stored as `self.sources`.

        :raises ValueError: If self.uploader_info is set to `auto` but no `uploader_name` is specified.
        :raises ValueError: If self.uploader_info is not set to `auto` or `manual`.
        """
        self.client = Client(self.address, self.cluster)
        if self.uploader_info == "auto":
            if not self.uploader_name:
                raise ValueError(
                    "uploader_name can not be empty if uploader_info is 'auto'"
                )
            self._get_uploader_info(self.uploader_name)
        elif self.uploader_info == "manual":
            pass
        else:
            raise ValueError(
                f"uploader_info must be one of 'auto' or 'manual', but was {self.uploader_info}"
            )

        self.sources = self._list_all_sources()

    @property
    def need_targets(self):
        """Compute if targets have to be downloaded.

        :return: Whether targets (or labels) should be downloaded
        :rtype: bool
        """
        return self.target_prefix and not self.autoencoding

    def __len__(self):
        length = int(np.floor(self.num_samples / self.batch_size))
        return length

    def __iter__(self):

        if self.sources:
            self.update_data()
            # Generate data
            if len(self) < 1:
                msg = "Not enough samples in generator for one batch. "
                msg += "Please run init_samples() or initialize generator with init_samples=True"
                raise ValueError(msg)

            for index in range(len(self)):
                indices = self.indices[
                    index * self.batch_size : (index + 1) * self.batch_size
                ]

                x, y = self.__data_generation(indices)

                if y is not None:
                    yield x, y
                else:
                    yield x

    def init_samples(self, sources=None):
        """Initialize samples (and targets, if needed).

        This function will not return until samples have been downloaded
        from all sources.

        :param sources: List of sources as defined in `init_sources`, defaults to None,
                        in which case sources will be initialized, unless `self.sources`
                        is already set
        :type sources: list[tuple], optional
        """
        self.autoencoding = self.sample_prefix == self.target_prefix

        if sources is not None:
            self.sources = sources

        if self.sources is None:
            self.sources = self._list_all_sources()

        self.log("Generator initialization complete")

        self._update_samples_and_targets()
        if self.shuffle:
            np.random.shuffle(self.indices)

    def _data_exists(self, batch_name, target_name):

        if self.need_targets:
            return self.client.tensor_exists(batch_name) and self.client.tensor_exists(
                target_name
            )
        else:
            return self.client.tensor_exists(batch_name)

    def _add_samples(self, batch_name, target_name):
        if self.samples is None:
            self.samples = self.client.get_tensor(batch_name)
            if self.need_targets:
                self.targets = self.client.get_tensor(target_name)
        else:
            self.samples = np.concatenate(
                (self.samples, self.client.get_tensor(batch_name))
            )
            if self.need_targets:
                self.targets = np.concatenate(
                    (self.targets, self.client.get_tensor(target_name))
                )

        self.num_samples = self.samples.shape[0]
        self.indices = np.arange(self.num_samples)
        self.log("Success!")
        self.log(f"New dataset size: {self.num_samples}, batches: {len(self)}")

    def _get_uploader_info(self, uploader_name):
        dataset_name = form_name(uploader_name, "info")
        self.log(f"Uploader dataset name: {dataset_name}")

        ds_exists = False
        try:
            ds_exists = self.client.dataset_exists(dataset_name)
        # As long as required SmartRedis version is not 0.3 we
        # need a workaround for the missing function
        except AttributeError:
            try:
                uploaders = environ["SSKEYIN"].split(",")
                for uploader in uploaders:
                    if self.client.key_exists(uploader + "." + dataset_name):
                        ds_exists = True
            except KeyError:
                msg = "Uploader must be launched with SmartSim and added to incoming entity, "
                msg += "when setting uploader_info to 'auto'"
                raise SmartSimError(msg)

        trials = 6
        while not ds_exists:
            trials -= 1
            if trials == 0:
                raise SmartSimError("Could not find uploader dataset")
            time.sleep(5)
            try:
                ds_exists = self.client.dataset_exists(dataset_name)
            except AttributeError:
                try:
                    uploaders = environ["SSKEYIN"].split(",")
                    for uploader in uploaders:
                        if self.client.key_exists(uploader + "." + dataset_name):
                            ds_exists = True
                except KeyError:
                    msg = "Uploader must be launched with SmartSim and added to incoming entity, "
                    msg += "when setting uploader_info to 'auto'"
                    raise SmartSimError(msg)

        uploader_info = self.client.get_dataset(dataset_name)
        self.sample_prefix = uploader_info.get_meta_strings("sample_prefix")[0]
        self.log(f"Uploader sample prefix: {self.sample_prefix}")

        try:
            self.target_prefix = uploader_info.get_meta_strings("target_prefix")[0]
        except:
            self.target_prefix = None
        self.log(f"Uploader target prefix: {self.target_prefix}")

        try:
            self.producer_prefixes = uploader_info.get_meta_strings("producer_prefixes")
        except:
            self.producer_prefixes = [""]
        self.log(f"Uploader producer prefixes: {self.producer_prefixes}")

        try:
            self.num_classes = uploader_info.get_meta_scalars("num_classes")[0]
        except:
            self.num_classes = None
        self.log(f"Uploader num classes: {self.num_classes}")

        try:
            num_ranks = uploader_info.get_meta_scalars("num_ranks")[0]
            self.sub_indices = [str(rank) for rank in range(num_ranks)]
        except:
            self.sub_indices = None
        self.log(f"Uploader sub-indices: {self.sub_indices}")

    def _update_samples_and_targets(self):
        for source in self.sources:
            entity = source[0]
            sub_index = source[1]
            self.client.set_data_source(entity)
            batch_name = form_name(self.sample_prefix, sub_index)
            if self.need_targets:
                target_name = form_name(self.target_prefix, sub_index)
            else:
                target_name = None

            self.log(f"Retrieving {batch_name} from {entity}")
            trials = 6
            while not self._data_exists(batch_name, target_name):
                trials -= 1
                if trials == 0:
                    raise SmartSimError(
                        f"Could not retrieve batch {batch_name} from entity {entity}"
                    )
                time.sleep(5)

            self._add_samples(batch_name, target_name)

    def update_data(self):
        self._update_samples_and_targets()

    def __data_generation(self, indices):
        # Initialization
        x = self.samples[indices]

        if self.need_targets:
            y = self.targets[indices]

        elif self.autoencoding:
            y = x
        else:
            y = None

        return x, y

    def __len__(self):
        length = int(np.floor(self.num_samples / self.batch_size))
        return length


class DynamicDataDownloader(StaticDataDownloader):
    """A class to download batches from the DB as they are produced.

    By default, the DynamicDataDownloader has to be created in a process
    launched through SmartSim, with sample producers listed as incoming
    entities.

    All details about the batches must be defined in
    the constructor; two mechanisms are available, `manual` and
    `auto`.

     - When specifying `auto`, the user must also specify
      `uploader_name`. DynamicDataDownloader will get all needed information
      from the database (this expects a Dataset like the one created
      by TrainingDataUploader to be available and stored as `uploader_name`
      on the DB).

     - When specifying `manual`, the user must also specify details
       of batch naming. Specifically, for each incoming entity with
       a name starting with an element of `producer_prefixes`,
       DynamicDataDownloader will query the DB
       for all batches named <sample_prefix>_<sub_index>_<iteration> for all indices
       in `sub_indices` if supplied, and, if
       `target_prefix` is supplied, it will also query for all targets
       named <target_prefix>.<sub_index>.<iteration>. If `producer_prefixes` is
       None, then all incoming entities will be treated as producers,
       and for each one, the corresponding batches will be downloaded.

    The flag `init_samples` defines whether sources (the list of batches
    to be fetched) and samples (the actual data) should automatically
    be set up in the costructor.

    If the user needs to modify the list of sources, then `init_samples=False`
    has to be set. In that case, to set up a `DynamicDataDownloader`, the user has to call
    `init_sources()` (which initializes the list of sources and the SmartRedis client)
    and `init_samples()`.  After `init_sources()` is called,
    a list of data sources is populated, representing the batches which
    will be downloaded. See `init_sources()`

    Each source is represented as a tuple `(producer_name, sub_index, iteration)`.
    Before `init_samples()` is called, the user can modify the list.
    Once `init_samples()` is called, all data is downloaded and batches
    can be obtained with iter().

    After initialization, samples and targets can be updated calling `update_data()`,
    which shuffles the available samples, if `shuffle` is set to ``True`` at initialization.

    :param batch_size: Size of batches obtained with __iter__
    :type batch_size: int
    :param shuffle: whether order of samples has to be shuffled when calling `update_data`
    :type shuffle: bool
    :param uploader_info: Set to `auto` uploader information has to be downloaded from DB,
                          or to `manual` if it is provided by the user
    :type uploader_info: str
    :param uploader_name: Name of uploader info dataset, only used if `uploader_info` is `auto`
    :type uploader_name: str
    :param sample_prefix: prefix of keys representing batches
    :type sample_prefix: str
    :param target_prefix: prefix of keys representing targets
    :type target_prefix: str
    :param uploader_ranks: Number of processes every uploader runs on (e.g, if each
                           rank in an MPI simulation is uploading its own batches,
                           this will be the MPI comm world size of the simulation).
    :type uploader_ranks: int
    :param num_classes: Number of classes of targets, if categorical
    :type num_classes: int
    :param producer_prefixes: Prefixes of processes which will be producing batches.
                              This can be useful in case the consumer processes also
                              have other incoming entities.
    :type producer_prefixes: str
    :param cluster: Whether the Orchestrator is being run as a cluster
    :type cluster: bool
    :param address: Address of Redis DB as <ip_address>:<port>
    :type address: str
    :param replica_rank: When StaticDataDownloader is used in a distributed setting, indicates
                         the rank of this replica
    :type replica_rank: int
    :param num_replicas: When ContinuousBatchDownlaoder is used in a distributed setting,
                         indicates the total number of replicas (ranks)
    :type num_replicas: int
    :param verbose: Whether log messages should be printed
    :type verbose: bool
    :param init_samples: whether samples should be initialized in the constructor
    :type init_samples: bool
    """

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

    def _list_all_sources(self):
        sources = super()._list_all_sources()
        # Append the batch index to each source
        for source in sources:
            source.append(0)
        return sources

    def _update_samples_and_targets(self):
        for source in self.sources:
            entity = source[0]
            sub_index = source[1]
            index = source[2]
            self.client.set_data_source(entity)
            batch_name = form_name(self.sample_prefix, index, sub_index)
            if self.need_targets:
                target_name = form_name(self.target_prefix, index, sub_index)
            else:
                target_name = None

            self.log(f"Retrieving {batch_name} from {entity}")
            # Poll next batch based on index, if available: retrieve it, update index and loop
            while self._data_exists(batch_name, target_name):
                self._add_samples(batch_name, target_name)
                source[2] += 1
                index = source[2]
                batch_name = form_name(self.sample_prefix, index, sub_index)
                if self.need_targets:
                    target_name = form_name(self.target_prefix, index, sub_index)

                self.log(f"Retrieving {batch_name}...")

    def update_data(self):
        """Update data.

        Fetch new batches (if available) from the DB. Also shuffle
        list of samples if `self.shuffle` is set to ``True``.
        """
        self._update_samples_and_targets()
        if self.shuffle:
            np.random.shuffle(self.indices)

    def init_samples(self, sources=None):
        """Initialize samples (and targets, if needed).

        This function will not return until at least one batch worth of data
        has been downloaded.

        :param sources: List of sources as defined in `init_sources`, defaults to None,
                        in which case sources will be initialized, unless `self.sources`
                        is already set
        :type sources: list[tuple], optional
        """
        self.autoencoding = self.sample_prefix == self.target_prefix

        if sources is not None:
            self.sources = sources

        if self.sources is None:
            self.sources = self._list_all_sources()

        if self.sources:

            while len(self) < 1:
                self._update_samples_and_targets()
                trials = 6
                if len(self) < 1:
                    trials -= 1
                    if trials == 0:
                        raise SmartSimError("Could not find samples")
                    time.sleep(5)
            self.log("Generator initialization complete")
        else:
            self.log(
                "Generator has no associated sources, this can happen if the number of "
                "loader workers is larger than the number of available sources."
            )



import numpy as np
import tensorflow.keras as keras

from smartsim.ml import DynamicDataDownloader, StaticDataDownloader


class StaticDataGenerator(StaticDataDownloader, keras.utils.Sequence):
    """A class to download a dataset from the DB.

    Details about parameters and features of this class can be found
    in the documentation of ``StaticDataDownloader``, of which it is just
    a TensorFlow-specialized sub-class.
    """

    def __init__(self, **kwargs):
        StaticDataDownloader.__init__(self, **kwargs)

    def __getitem__(self, index):
        if len(self) < 1:
            msg = "Not enough samples in generator for one batch. "
            msg += "Please run init_samples() or initialize generator with init_samples=True"
            raise ValueError(msg)
        # Generate indices of the batch
        indices = self.indices[index * self.batch_size : (index + 1) * self.batch_size]

        # Generate data
        x, y = self.__data_generation(indices)

        if y is not None:
            return x, y
        else:
            return x

    def on_epoch_end(self):
        """Callback called at the end of each training epoch

        If `self.shuffle` is set to `True`, data is shuffled.
        """
        if self.shuffle:
            np.random.shuffle(self.indices)

    def __data_generation(self, indices):
        # Initialization
        x = self.samples[indices]

        if self.need_targets:
            y = self.targets[indices]
            if self.num_classes is not None:
                y = keras.utils.to_categorical(y, num_classes=self.num_classes)
        elif self.autoencoding:
            y = x
        else:
            y = None

        return x, y


class DynamicDataGenerator(DynamicDataDownloader, StaticDataGenerator):
    """A class to download batches from the DB.

    Details about parameters and features of this class can be found
    in the documentation of ``DynamicDataDownloader``, of which it is just
    a TensorFlow-specialized sub-class.
    """

    def __init__(self, **kwargs):
        StaticDataGenerator.__init__(self, **kwargs)

    def on_epoch_end(self):
        """Callback called at the end of each training epoch

        Update data (the DB is queried for new batches) and
        if `self.shuffle` is set to `True`, data is also shuffled.
        """
        self.update_data()
        super().on_epoch_end()



# BSD 2-Clause License
#
# Copyright (c) 2021, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import argparse
from typing import List
from subprocess import PIPE, STDOUT, Popen

from smartsim._core.utils.network import current_ip
from smartsim.exp.ray import parse_ray_head_node_address

def main(network_interface: str,
         port: int,
         is_head: bool,
         password: str,
         ray_exe: str,
         ray_args: List[str],
         dash_port: str,
         head_log: str,
         ):

    ip_address = current_ip(network_interface)

    cliargs = [
        ray_exe,
        "start",
        "--head"
        if is_head
        else f"--address={parse_ray_head_node_address(head_log)}:{port}",
        "--block",
        f"--node-ip-address={ip_address}",
    ]

    if ray_args:
        cliargs += ray_args
        if is_head and not any(
            [arg.startswith("--dashboard-host") for arg in ray_args]
        ):
            cliargs += [f"--dashboard-host={ip_address}"]

    if password:
        cliargs += [f"--redis-password={password}"]

    if is_head:
        cliargs += [f"--port={port}", f"--dashboard-port={dash_port}"]


    cmd = " ".join(cliargs)
    print(f"Ray Command: {cmd}")

    p = Popen(cliargs, stdout=PIPE, stderr=STDOUT)

    for line in iter(p.stdout.readline, b""):
        print(line.decode("utf-8").rstrip(), flush=True)


if __name__ == "__main__":

    os.environ["PYTHONUNBUFFERED"] = "1"

    parser = argparse.ArgumentParser(
        prefix_chars="+", description="SmartSim Ray head launcher"
    )
    parser.add_argument(
        "+port", type=int, help="Port used by Ray to start the Redis server at"
    )
    parser.add_argument("+head", action="store_true")
    parser.add_argument("+redis-password", type=str, help="Password of Redis cluster")
    parser.add_argument(
        "+ray-args", action="append", help="Additional arguments to start Ray"
    )
    parser.add_argument("+dashboard-port", type=str, help="Ray dashboard port")
    parser.add_argument("+ray-exe", type=str, help="Ray executable", default="ray")
    parser.add_argument("+ifname", type=str, help="Interface name", default="lo")
    parser.add_argument("+head-log", type=str, help="Head node log")
    args = parser.parse_args()

    if not args.head and not args.head_log:
        raise argparse.ArgumentError(
            "Ray starter needs +head or +head-log to start head or worker nodes respectively"
        )

    main(args.ifname,
         args.port,
         args.head,
         args.redis_password,
         args.ray_exe,
         args.ray_args,
         args.dashboard_port,
         args.head_log)


# BSD 2-Clause License
#
# Copyright (c) 2021-2022, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import re
import time as _time
import uuid
import sys

from ..._core.utils import init_default
from ..._core.utils.helpers import expand_exe_path
from ...entity import EntityList, SmartSimEntity
from ...error import SmartSimError, SSUnsupportedError
from ...log import get_logger
from ...settings import settings

logger = get_logger(__name__)


class RayCluster(EntityList):
    """Entity used to run a Ray cluster on a given number of hosts. One Ray node is
    launched on each host, and the first host is used to launch the head node.

    :param name: The name of the entity.
    :type name: str
    :param path: path to output, error, and configuration files
    :type path: str
    :param ray_port: Port at which the head node will be running.
    :type ray_port: int
    :param ray_args: Arguments to be passed to Ray executable.
    :type ray_args: dict[str,str]
    :param num_nodes: Number of hosts, includes 1 head node and all worker nodes.
    :type num_nodes: int
    :param run_args: Arguments to pass to launcher to specify details such as partition or time.
    :type run_args: dict[str,str]
    :param batch_args: Additional batch arguments passed to launcher when running batch jobs.
    :type batch_args: dict[str,str]
    :param launcher: Name of launcher to use for starting the cluster.
    :type launcher: str
    :param interface: Name of network interface the cluster nodes should bind to.
    :type interface: str
    :param alloc: ID of allocation to run on, if obtained with ``smartsim.slurm.get_allocation``
    :type alloc: int
    :param batch: Whether cluster should be launched as batch file, ignored when ``launcher`` is `local`
    :type batch: bool
    :param time: The walltime the cluster will be running for
    :type time: str
    :param run_command: specify launch binary, defaults to automatic selection.
    :type run_command: str
    :param hosts: specify hosts to launch on, defaults to None. Optional if not launching with OpenMPI.
    :type hosts: str, list[str]
    :param password: Password to use for Redis server, which is passed as `--redis_password` to `ray start`.
                     Can be set to
                     - `auto`: a strong password will be generated internally
                     - a string: it will be used as password
                     - `None`: the default Ray password will be used.
                     Defaults to `auto`
    :type password: str
    """

    def __init__(
        self,
        name,
        path=os.getcwd(),
        ray_port=6789,
        ray_args=None,
        num_nodes=1,
        run_args=None,
        batch_args=None,
        launcher="local",
        batch=False,
        time="01:00:00",
        interface="ipogif0",
        alloc=None,
        run_command=None,
        host_list=None,
        password="auto",
        **kwargs,
    ):
        launcher = launcher.lower()
        supported_launchers = ["slurm", "pbs", "cobalt"]
        if launcher not in supported_launchers:
            raise SSUnsupportedError(
                "The supported launchers for RayCluster are",
                *[f"{launcher_name}," for launcher_name in supported_launchers],
                f"but {launcher} was provided.",
            )

        if password:
            if password == "auto":
                self._ray_password = str(uuid.uuid4())
            else:
                self._ray_password = password
        else:
            self._ray_password = None

        if num_nodes < 1:
            raise ValueError("Number of nodes must be larger than 0.")

        self.alloc = None
        self.batch_settings = None
        self._hosts = None

        run_args = init_default({}, run_args, dict)
        batch_args = init_default({}, batch_args, dict)
        ray_args = init_default({}, ray_args, dict)

        self._ray_args = ray_args
        super().__init__(
            name=name,
            path=path,
            ray_args=ray_args,
            run_args=run_args,
            ray_port=ray_port,
            launcher=launcher,
            interface=interface,
            alloc=alloc,
            num_nodes=num_nodes,
            run_command=run_command if run_command else "auto",
            host_list=host_list,
            **kwargs,
        )
        if batch:
            self.batch_settings = settings.create_batch_settings(
                launcher=launcher,
                nodes=num_nodes,
                time=time,
                batch_args=batch_args,
                **kwargs,
            )
        self.ray_head_address = None

        if host_list:
            self.set_hosts(host_list=host_list, launcher=launcher)

    @property
    def batch(self):
        try:
            if self.batch_settings:
                return True
            return False
        except AttributeError:
            return False

    def set_hosts(self, host_list, launcher):
        """Specify the hosts for the ``RayCluster`` to launch on. This is
        optional, unless ``run_command`` is `mpirun`.

        :param host_list: list of hosts (compute node names)
        :type host_list: str | list[str]
        :raises TypeError: if wrong type
        """
        if isinstance(host_list, str):
            host_list = [host_list.strip()]
        if not isinstance(host_list, list):
            raise TypeError("host_list argument must be a list of strings")
        if not all([isinstance(host, str) for host in host_list]):
            raise TypeError("host_list argument must be list of strings")
        # TODO check length
        if self.batch:
            self.batch_settings.set_hostlist(host_list)
        for host, node in zip(host_list, self.entities):
            # Aprun doesn't like settings hosts in batch launch
            if launcher == "pbs" or launcher == "cobalt":
                if not self.batch:
                    node.run_settings.set_hostlist([host])
            else:
                node.run_settings.set_hostlist([host])

    def _initialize_entities(self, **kwargs):

        ray_port = kwargs.get("ray_port", 6789)
        launcher = kwargs.get("launcher", "slurm")
        ray_args = kwargs.get("ray_args", None)
        run_args = kwargs.get("run_args", None)
        interface = kwargs.get("interface", "ipogif0")
        num_nodes = kwargs.get("num_nodes", 0)
        alloc = kwargs.get("alloc", None)
        run_command = kwargs.get("run_command", None)

        ray_head = RayHead(
            name="ray_head",
            path=self.path,
            ray_password=self._ray_password,
            ray_port=ray_port,
            launcher=launcher,
            run_args=run_args.copy(),
            ray_args=ray_args.copy(),
            interface=interface,
            run_command=run_command,
            alloc=alloc,
        )

        self.entities.append(ray_head)

        for worker_id in range(num_nodes - 1):
            worker_model = RayWorker(
                name=f"ray_worker_{worker_id}",
                path=self.path,
                run_args=run_args.copy(),
                ray_port=ray_port,
                ray_password=self._ray_password,
                ray_args=ray_args.copy(),
                interface=interface,
                run_command=run_command,
                launcher=launcher,
                alloc=alloc,
            )
            self.entities.append(worker_model)

    def get_head_address(self):
        """Return address of head node

        If address has not been initialized, returns None

        :returns: Address of head node
        :rtype: str
        """
        if not self.ray_head_address:
            self.ray_head_address = parse_ray_head_node_address(
                os.path.join(self.entities[0].path, self.entities[0].name + ".out")
            )
        return self.ray_head_address

    def get_dashboard_address(self):
        """Returns dashboard address

        The format is <head_ip>:<dashboard_port>

        :returns: Dashboard address
        :rtype: str
        """
        return self.get_head_address() + ":" + str(self.entities[0].dashboard_port)

    def _update_workers(self):
        """Update worker args before launching them."""
        for worker in range(1, len(self.entities)):
            self.entities[worker].set_head_log(
                f"{os.path.join(self.entities[0].path, self.entities[0].name)}.out"
            )


def find_ray_exe():
    """Find ray executable in current path."""
    # TODO add this to CONFIG?
    try:
        return expand_exe_path("ray")
    except (TypeError, FileNotFoundError):
        raise SmartSimError("Could not find ray executable")


def parse_ray_head_node_address(head_log):
    """Get the ray head node host address from the log file produced
    by the head process.

    :param head_log: full path to log file of head node
    :return: address of the head host
    :rtype: str
    """

    max_attempts = 24
    attempts = 0
    while not os.path.isfile(head_log):
        _time.sleep(5)
        attempts += 1
        if attempts == max_attempts:
            raise RuntimeError(f"Could not find Ray cluster head log file {head_log}")

    attempts = 0
    head_ip = None
    while head_ip is None:
        _time.sleep(5)
        with open(head_log) as fp:
            line = fp.readline()
            while line:
                plain_line = re.sub("\033\\[([0-9]+)(;[0-9]+)*m", "", line)
                if "Local node IP:" in plain_line:
                    matches = re.search(r"(?<=Local node IP: ).*", plain_line)
                    head_ip = matches.group()
                    break
                line = fp.readline()
        attempts += 1
        if attempts == max_attempts:
            raise RuntimeError(
                f"Could not find Ray cluster head address in log file {head_log}."
            )

    return head_ip


class RayHead(SmartSimEntity):
    def __init__(
        self,
        name,
        path,
        ray_password,
        ray_port=6789,
        run_args=None,
        ray_args=None,
        launcher="slurm",
        interface="ipogif0",
        run_command=None,
        alloc=None,
        dash_port=8265,
        **kwargs,
    ):
        self.dashboard_port = dash_port
        self.batch_settings = None
        self.files = None

        run_args = init_default({}, run_args, dict)
        ray_args = init_default({}, ray_args, dict)

        ray_exe_args = self._build_ray_exe_args(
            ray_port, ray_password, interface, ray_args
        )

        run_settings = settings.create_run_settings(
            launcher=launcher,
            exe="python",
            exe_args=ray_exe_args,
            run_args=run_args,
            run_command=run_command if run_command else "auto",
            alloc=alloc,
            **kwargs,
        )

        run_settings.set_tasks_per_node(1)
        run_settings.set_tasks(1)

        super().__init__(name, path, run_settings)

    def _build_ray_exe_args(self, ray_port, ray_password, interface, ray_args):

        # python script that launches ray head node
        ray_starter_args = [
            "-m",
            "smartsim._core.entrypoints.ray",
            f"+port={ray_port}",
            f"+ifname={interface}",
            f"+ray-exe={find_ray_exe()}",
            f"+head",
        ]

        if ray_password:
            ray_starter_args += [f"+redis-password={ray_password}"]

        if "dashboard-port" in ray_args:
            self.dashboard_port = int(ray_args["dashboard-port"])
        ray_starter_args += [f"+dashboard-port={self.dashboard_port}"]

        used = ["block", "redis-password", "start", "head", "port", "dashboard-port"]
        extra_ray_args = []
        for key, value in ray_args.items():
            if key not in used:
                extra_ray_args += [f"+ray-args=--{key}={value}"]
        ray_starter_args += extra_ray_args

        return " ".join(ray_starter_args)


class RayWorker(SmartSimEntity):
    def __init__(
        self,
        name,
        path,
        ray_password,
        ray_port,
        run_args=None,
        ray_args=None,
        interface="ipogif0",
        launcher="slurm",
        run_command=None,
        alloc=None,
        **kwargs,
    ):

        self.batch_settings = None
        self.files = None

        run_args = init_default({}, run_args, dict)
        ray_args = init_default({}, ray_args, dict)

        ray_exe_args = self._build_ray_exe_args(
            ray_password, ray_args, ray_port, interface
        )

        run_settings = settings.create_run_settings(
            launcher=launcher,
            exe=sys.executable,
            exe_args=ray_exe_args,
            run_args=run_args,
            run_command=run_command,
            alloc=alloc,
            **kwargs,
        )

        run_settings.set_tasks_per_node(1)
        run_settings.set_tasks(1)

        super().__init__(name, path, run_settings)

    @property
    def batch(self):
        return False

    def set_head_log(self, head_log):
        """Set head log file (with full path)

        The head log file is used by the worker to discover
        the head IP address. This function is called by
        RayCluster before the cluster is launched.
        """
        self.run_settings.add_exe_args([f"+head-log={head_log}"])

    def _build_ray_exe_args(self, ray_password, ray_args, ray_port, interface):

        # python script that launches ray  node
        ray_starter_args = [
            "-m",
            "smartsim._core.entrypoints.ray",
            f"+ray-exe={find_ray_exe()}",
            f"+port={ray_port}",
            f"+ifname={interface}",
        ]
        if ray_password:
            ray_starter_args += [f"+redis-password={ray_password}"]

        used = [
            "block",
            "redis-password",
            "start",
            "head",
            "port",
            "dashboard-port",
            "dashboard-host",
        ]
        extra_ray_args = []
        for key, value in ray_args.items():
            if key not in used:
                extra_ray_args += [f"+ray-args=--{key}={value}"]
        ray_starter_args += extra_ray_args

        return " ".join(ray_starter_args)



# BSD 2-Clause License
#
# Copyright (c) 2021-2022, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

from .._core.utils.helpers import (
    expand_exe_path,
    init_default,
    is_valid_cmd,
    fmt_dict
)
from ..log import get_logger

logger = get_logger(__name__)


class RunSettings:
    def __init__(
        self, exe, exe_args=None, run_command="", run_args=None, env_vars=None, **kwargs
    ):
        """Run parameters for a ``Model``

        The base ``RunSettings`` class should only be used with the `local`
        launcher on single node, workstations, or laptops.

        If no ``run_command`` is specified, the executable will be launched
        locally.

        ``run_args`` passed as a dict will be interpreted literally for
        local ``RunSettings`` and added directly to the ``run_command``
        e.g. run_args = {"-np": 2} will be "-np 2"

        Example initialization

        .. highlight:: python
        .. code-block:: python

            rs = RunSettings("echo", "hello", "mpirun", run_args={"-np": "2"})

        :param exe: executable to run
        :type exe: str
        :param exe_args: executable arguments, defaults to None
        :type exe_args: str | list[str], optional
        :param run_command: launch binary (e.g. "srun"), defaults to empty str
        :type run_command: str, optional
        :param run_args: arguments for run command (e.g. `-np` for `mpiexec`), defaults to None
        :type run_args: dict[str, str], optional
        :param env_vars: environment vars to launch job with, defaults to None
        :type env_vars: dict[str, str], optional
        """
        self.exe = [expand_exe_path(exe)]
        self.exe_args = self._set_exe_args(exe_args)
        self.run_args = init_default({}, run_args, dict)
        self.env_vars = init_default({}, env_vars, dict)
        self._run_command = run_command
        self.in_batch = False
        self.colocated_db_settings = None

    def set_tasks(self, tasks):
        """Set the number of tasks to launch

        :param tasks: number of tasks to launch
        :type tasks: int
        """
        raise NotImplementedError(
            f"Task specification not implemented for this RunSettings type: {type(self)}"
        )

    def set_tasks_per_node(self, tasks_per_node):
        """Set the number of tasks per node

        :param tasks_per_node: number of tasks to launch per node
        :type tasks_per_node: int
        """
        raise NotImplementedError(
            f"Task per node specification not implemented for this RunSettings type: {type(self)}"
        )

    def set_cpus_per_task(self, cpus_per_task):
        """Set the number of cpus per task

        :param cpus_per_task: number of cpus per task
        :type cpus_per_task: int
        """
        raise NotImplementedError(
            f"CPU per node specification not implemented for this RunSettings type: {type(self)}"
        )

    def set_hostlist(self, host_list):
        """Specify the hostlist for this job

        :param host_list: hosts to launch on
        :type host_list: str | list[str]
        """
        raise NotImplementedError(
            f"Host list specification not implemented for this RunSettings type: {type(self)}"
        )

    @property
    def run_command(self):
        """Return the launch binary used to launch the executable

        Attempt to expand the path to the executable if possible

        :returns: launch binary e.g. mpiexec
        :type: str | None
        """
        if self._run_command:
            if is_valid_cmd(self._run_command):
                # command is valid and will be expanded
                return expand_exe_path(self._run_command)
            # command is not valid, so return it as is
            # it may be on the compute nodes but not local machine
            return self._run_command
        # run without run command
        return None

    def update_env(self, env_vars):
        """Update the job environment variables

        :param env_vars: environment variables to update or add
        :type env_vars: dict[str, str]
        """
        self.env_vars.update(env_vars)

    def add_exe_args(self, args):
        """Add executable arguments to executable

        :param args: executable arguments
        :type args: str | list[str]
        :raises TypeError: if exe args are not strings
        """
        if isinstance(args, str):
            args = args.split()
        for arg in args:
            if not isinstance(arg, str):
                raise TypeError("Executable arguments should be a list of str")
            self.exe_args.append(arg)

    def _set_exe_args(self, exe_args):
        if exe_args:
            if isinstance(exe_args, str):
                return exe_args.split()
            if isinstance(exe_args, list):
                plain_type = all([isinstance(arg, (str)) for arg in exe_args])
                if not plain_type:
                    nested_type = all(
                        [
                            all([isinstance(arg, (str)) for arg in exe_args_list])
                            for exe_args_list in exe_args
                        ]
                    )
                    if not nested_type:
                        raise TypeError(
                            "Executable arguments were not list of str or str"
                        )
                    else:
                        return exe_args
                return exe_args
            raise TypeError("Executable arguments were not list of str or str")
        else:
            return []

    def format_run_args(self):
        """Return formatted run arguments

        For ``RunSettings``, the run arguments are passed
        literally with no formatting.

        :return: list run arguments for these settings
        :rtype: list[str]
        """
        formatted = []
        for arg, value in self.run_args.items():
            formatted.append(arg)
            formatted.append(str(value))
        return formatted

    def __str__(self): # pragma: no-cover
        string = f"Executable: {self.exe[0]}\n"
        string += f"Executable Arguments: {' '.join((self.exe_args))}"
        if self.run_command:
            string += f"\nRun Command: {self._run_command}"
        if self.run_args:
            string += f"\nRun Arguments:\n{fmt_dict(self.run_args)}"
        if self.colocated_db_settings:
            string += "\nCo-located Database: True"
        return string


class BatchSettings:
    def __init__(self, batch_cmd, batch_args=None, **kwargs):
        self._batch_cmd = batch_cmd
        self.batch_args = init_default({}, batch_args, dict)
        self._preamble = []
        self.set_nodes(kwargs.get("nodes", None))
        self.set_walltime(kwargs.get("time", None))
        self.set_queue(kwargs.get("queue", None))
        self.set_account(kwargs.get("account", None))

    @property
    def batch_cmd(self):
        """Return the batch command

        Tests to see if we can expand the batch command
        path. If we can, then returns the expanded batch
        command. If we cannot, returns the batch command as is.

        :returns: batch command
        :type: str
        """
        if is_valid_cmd(self._batch_cmd):
            return expand_exe_path(self._batch_cmd)
        else:
            return self._batch_cmd

    def set_nodes(self, num_nodes):
        raise NotImplementedError

    def set_hostlist(self, host_list):
        raise NotImplementedError

    def set_queue(self, queue):
        raise NotImplementedError

    def set_walltime(self, walltime):
        raise NotImplementedError

    def set_account(self, account):
        raise NotImplementedError

    def format_batch_args(self):
        raise NotImplementedError

    def set_batch_command(self, command):
        """Set the command used to launch the batch e.g. ``sbatch``

        :param command: batch command
        :type command: str
        """
        self._batch_cmd = command

    def add_preamble(self, lines):
        """Add lines to the batch file preamble. The lines are just
        written (unmodified) at the beginning of the batch file
        (after the WLM directives) and can be used to e.g.
        start virtual environments before running the executables.

        :param line: lines to add to preamble.
        :type line: str or list[str]
        """
        if isinstance(lines, str):
            self._preamble += [lines]
        elif isinstance(lines, list):
            self._preamble += lines
        else:
            raise TypeError("Expected str or List[str] for lines argument")

    def __str__(self): # pragma: no-cover
        string = f"Batch Command: {self._batch_cmd}"
        if self.batch_args:
            string += f"\nBatch arguments:\n{fmt_dict(self.batch_args)}"
        return string



# BSD 2-Clause License
#
# Copyright (c) 2021, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import signal
import psutil
import argparse
from typing import List
from subprocess import PIPE, STDOUT

from smartsim.log import get_logger
from smartsim.error import SSInternalError
from smartsim._core.utils.network import current_ip

logger = get_logger(__name__)

"""
Redis/KeyDB entrypoint script
"""

DBPID = None

# kill is not catchable
SIGNALS = [
    signal.SIGINT,
    signal.SIGQUIT,
    signal.SIGTERM,
    signal.SIGABRT
    ]

def handle_signal(signo, frame):
    cleanup()


def main(network_interface: str, command: List[str]):
    global DBPID

    try:

        ip_address = current_ip(network_interface)
        cmd = command + [f"--bind {ip_address}"]

        print("-" * 10, "  Running  Command  ", "-" * 10, "\n", flush=True)
        print(f"COMMAND: {' '.join(cmd)}\n", flush=True)
        print(f"IPADDRESS: {ip_address}\n", flush=True)
        print(f"NETWORK: {network_interface}\n", flush=True)
        print("-" * 30, "\n\n", flush=True)

        print("-" * 10, "  Output  ", "-" * 10, "\n\n", flush=True)

        p = psutil.Popen(cmd, stdout=PIPE, stderr=STDOUT)
        DBPID = p.pid

        for line in iter(p.stdout.readline, b""):
            print(line.decode("utf-8").rstrip(), flush=True)
    except Exception as e:
        cleanup()
        raise SSInternalError("Database process starter raised an exception") from e


def cleanup():
    global DBPID
    try:
        logger.debug("Cleaning up database instance")
        # attempt to stop the database process
        db_proc = psutil.Process(DBPID)
        db_proc.terminate()

    except psutil.NoSuchProcess:
        logger.warning("Couldn't find database process to kill.")

    except OSError as e:
        logger.warning(
            f"Failed to clean up database gracefully: {str(e)}"
        )


if __name__ == "__main__":

    os.environ["PYTHONUNBUFFERED"] = "1"

    parser = argparse.ArgumentParser(
        prefix_chars="+", description="SmartSim Process Launcher"
    )
    parser.add_argument("+ifname", type=str, help="Network Interface name", default="lo")
    parser.add_argument("+command", nargs="+", help="Command to run")
    args = parser.parse_args()

    # make sure to register the cleanup before the start
    # the proecss so our signaller will be able to stop
    # the database process.
    for sig in SIGNALS:
        signal.signal(sig, handle_signal)

    main(args.ifname, args.command)


# BSD 2-Clause License
#
# Copyright (c) 2021-2022, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import logging
import time

import redis
from rediscluster import RedisCluster
from rediscluster.exceptions import ClusterDownError, RedisClusterException

logging.getLogger("rediscluster").setLevel(logging.WARNING)

from ...error import SSInternalError
from ...log import get_logger
from ..config import CONFIG
from ..launcher.util.shell import execute_cmd
from .network import get_ip_from_host

logger = get_logger(__name__)


def create_cluster(hosts, ports):  # cov-wlm
    """Connect launched cluster instances.

    Should only be used in the case where cluster initialization
    needs to occur manually which is not often.

    :param hosts: List of hostnames to connect to
    :type hosts: List[str]
    :param ports: List of ports for each hostname
    :type ports: List[int]
    :raises SmartSimError: if cluster creation fails
    """
    ip_list = []
    for host in hosts:
        ip = get_ip_from_host(host)
        for port in ports:
            address = ":".join((ip, str(port) + " "))
            ip_list.append(address)

    # call cluster command
    redis_cli = CONFIG.redis_cli
    cmd = [redis_cli, "--cluster", "create"]
    cmd += ip_list
    cmd += ["--cluster-replicas", "0"]
    returncode, out, err = execute_cmd(cmd, proc_input="yes", shell=False)

    if returncode != 0:
        logger.error(out)
        logger.error(err)
        raise SSInternalError("Database '--cluster create' command failed")
    logger.debug(out)


def check_cluster_status(hosts, ports, trials=10):  # cov-wlm
    """Check that a Redis/KeyDB cluster is up and running

    :param hosts: List of hostnames to connect to
    :type hosts: List[str]
    :param ports: List of ports for each hostname
    :type ports: List[int]
    :param trials: number of attempts to verify cluster status
    :type trials: int, optional

    :raises SmartSimError: If cluster status cannot be verified
    """
    host_list = []
    for host in hosts:
        for port in ports:
            host_dict = dict()
            host_dict["host"] = get_ip_from_host(host)
            host_dict["port"] = port
            host_list.append(host_dict)

    logger.debug("Beginning database cluster status check...")
    while trials > 0:
        # wait for cluster to spin up
        time.sleep(5)
        try:
            redis_tester = RedisCluster(startup_nodes=host_list)
            redis_tester.set("__test__", "__test__")
            redis_tester.delete("__test__")
            logger.debug("Cluster status verified")
            return
        except (ClusterDownError, RedisClusterException, redis.RedisError):
            logger.debug("Cluster still spinning up...")
            trials -= 1
    if trials == 0:
        raise SSInternalError("Cluster setup could not be verified")



import argparse
import sys
from pathlib import Path

import pkg_resources
from tabulate import tabulate

from smartsim._core._cli.utils import color_bool, pip_install
from smartsim._core._install import builder
from smartsim._core._install.buildenv import BuildEnv, SetupError, Version_, Versioner
from smartsim._core._install.builder import BuildError
from smartsim.log import get_logger

smart_logger_format = "[%(name)s] %(levelname)s %(message)s"
logger = get_logger("Smart", fmt=smart_logger_format)

# NOTE: all smartsim modules need full paths as the smart cli
#       may be installed into a different directory.


class Build:
    def __init__(self):
        parser = argparse.ArgumentParser()
        parser.add_argument(
            "-v",
            action="store_true",
            default=False,
            help="Enable verbose build process",
        )
        parser.add_argument(
            "--device",
            type=str,
            default="cpu",
            help="Device to build ML runtimes for (cpu || gpu)",
        )
        parser.add_argument(
            "--no_pt",
            action="store_true",
            default=False,
            help="Do not build PyTorch backend",
        )
        parser.add_argument(
            "--no_tf",
            action="store_true",
            default=False,
            help="Do not build TensorFlow backend",
        )
        parser.add_argument(
            "--onnx",
            action="store_true",
            default=False,
            help="Build ONNX backend (off by default)",
        )
        parser.add_argument(
            "--torch_dir",
            default=None,
            type=str,
            help="Path to custom <path>/torch/share/cmake/Torch/ directory (ONLY USE IF NEEDED)",
        )
        args = parser.parse_args(sys.argv[2:])
        self.verbose = args.v

        # torch and tf build by default
        pt = not args.no_pt
        tf = not args.no_tf
        onnx = args.onnx

        logger.info("Running SmartSim build process...")
        try:
            logger.info("Checking for build tools...")
            self.build_env = BuildEnv()

            if self.verbose:
                logger.info("Build Environment:")
                env = self.build_env.as_dict()
                print(tabulate(env, headers=env.keys(), tablefmt="github"), "\n")

            logger.info("Checking requested versions...")
            self.versions = Versioner()

            if self.verbose:
                logger.info("Version Information:")
                vers = self.versions.as_dict()
                print(tabulate(vers, headers=vers.keys(), tablefmt="github"), "\n")

            # REDIS
            self.build_redis()

            # REDISAI
            self.build_redis_ai(str(args.device), pt, tf, onnx, args.torch_dir)

        except (SetupError, BuildError) as e:
            logger.error(str(e))
            exit(1)

        logger.info("SmartSim build complete!")

    def build_redis(self):
        # check redis installation
        redis_builder = builder.RedisBuilder(
            self.build_env(), self.build_env.MALLOC, self.build_env.JOBS, self.verbose
        )

        if not redis_builder.is_built:
            logger.info(
                f"Building Redis version {self.versions.REDIS} from {self.versions.REDIS_URL}"
            )

            redis_builder.build_from_git(
                self.versions.REDIS_URL, self.versions.REDIS_BRANCH
            )
            redis_builder.cleanup()
        logger.info("Redis build complete!")

    def build_redis_ai(self, device, torch=True, tf=True, onnx=False, torch_dir=None):

        # make sure user isn't trying to do something silly on MacOS
        if self.build_env.PLATFORM == "darwin":
            if device == "gpu":
                logger.error("SmartSim does not support GPU on MacOS")
                exit(1)
            if onnx and self.versions.REDISAI < "1.2.6":
                logger.error("RedisAI < 1.2.6 does not support ONNX on MacOS")
                exit(1)
            if self.versions.REDISAI == "1.2.4" or self.versions.REDISAI == "1.2.5":
                logger.error("RedisAI support for MacOS is broken in 1.2.4 and 1.2.5")
                exit(1)

        # decide which runtimes to build
        print("\nML Backends Requested")
        print("-----------------------")
        print(f"    PyTorch {self.versions.TORCH}: {color_bool(torch)}")
        print(f"    TensorFlow {self.versions.TENSORFLOW}: {color_bool(tf)}")
        print(f"    ONNX {self.versions.ONNX}: {color_bool(onnx)}\n")
        print(f"Building for GPU support: {color_bool(device == 'gpu')}\n")

        # Check for onnx and tf in user python environemnt and prompt user
        # to download them if they are not installed. this should not break
        # the build however, as we use onnx and tf directly from RAI instead
        # of pip like we do PyTorch.
        if onnx:
            self.check_onnx_install()
        if tf:
            self.check_tf_install()

        cmd = []
        # TORCH
        if torch:
            if torch_dir:
                torch_dir = Path(torch_dir).resolve()
                if not torch_dir.is_dir():
                    # we will always be able to find a torch version downloaded by
                    # pip so if we can't find it we know the user suggested a torch
                    # installation path that doesn't exist
                    logger.error("Could not find requested user Torch installation")
                    exit(1)
            else:
                # install pytorch wheel, and get the path to the cmake dir
                # we will use in the RAI build
                self.install_torch(device=device)
                torch_dir = self.build_env.torch_cmake_path

        rai_builder = builder.RedisAIBuilder(
            build_env=self.build_env(),
            torch_dir=str(torch_dir),
            build_torch=torch,
            build_tf=tf,
            build_onnx=onnx,
            jobs=self.build_env.JOBS,
            verbose=self.verbose,
        )

        if rai_builder.is_built:
            logger.info("RedisAI installed. Run `smart clean` to remove.")
        else:
            # get the build environment, update with CUDNN env vars
            # if present and building for GPU, otherwise warn the user
            build_env = self.build_env()
            if device == "gpu":
                gpu_env = self.build_env.get_cudnn_env()
                cudnn_env_vars = [
                    "CUDNN_LIBRARY",
                    "CUDNN_INCLUDE_DIR",
                    "CUDNN_INCLUDE_PATH",
                    "CUDNN_LIBRARY_PATH",
                ]
                if not gpu_env:
                    logger.warning(
                        f"CUDNN environment variables not found.\n"
                        + f"Looked for {cudnn_env_vars}"
                    )
                else:
                    build_env.update(gpu_env)
            # update RAI build env with cudnn env vars
            rai_builder.env = build_env

            logger.info(
                f"Building RedisAI version {self.versions.REDISAI}"
                f" from {self.versions.REDISAI_URL}"
            )

            # NOTE: have the option to add other builds here in the future
            # like "from_tarball"
            rai_builder.build_from_git(
                self.versions.REDISAI_URL, self.versions.REDISAI_BRANCH, device
            )
            logger.info("ML Backends and RedisAI build complete!")

    def install_torch(self, device="cpu"):
        """Torch shared libraries installed by pip are used in the build
        for SmartSim backends so we download them here.
        """
        packages = []
        end_point = None
        if not self.build_env.check_installed("torch", self.versions.TORCH):
            # if we are on linux cpu, use the torch without CUDA
            if sys.platform == "linux" and device == "cpu":
                packages.append(f"torch=={self.versions.TORCH}+cpu")
                packages.append(f"torchvision=={self.versions.TORCHVISION}+cpu")
                end_point = "https://download.pytorch.org/whl/torch_stable.html"

            # otherwise just use the version downloaded by pip
            else:
                packages.append(f"torch=={self.versions.TORCH}")
                packages.append(f"torchvision=={self.versions.TORCHVISION}")

            pip_install(packages, end_point=end_point, verbose=self.verbose)

        # if torch already installed, check the versions to make sure correct
        # torch version is downloaded for that particular device
        else:
            installed = Version_(pkg_resources.get_distribution("torch").version)
            if device == "gpu":
                # if torch version is x.x.x+cpu
                if "cpu" in installed.patch:
                    msg = (
                        "Torch CPU is currently installed but torch GPU requested. Uninstall all torch packages "
                        + "and run the `smart build` command again to obtain Torch GPU libraries"
                    )
                    logger.warning(msg)

            if device == "cpu":
                # if torch version if x.x.x then we need to install the cpu version
                if "cpu" not in installed.patch and not self.build_env.is_macos():
                    msg = (
                        "Torch GPU installed in python environment but requested Torch CPU. "
                        + " Run `pip uninstall torch torchvision` and run `smart build` again"
                    )
                    logger.error(msg)  # error because this is usually fatal
            logger.info(f"Torch {self.versions.TORCH} installed in Python environment")

    def check_onnx_install(self):
        """Check Python environment for ONNX installation"""
        try:
            if not self.build_env.check_installed("onnx", self.versions.ONNX):
                msg = (
                    f"ONNX {self.versions.ONNX} not installed in python environment. "
                    + f"Consider installing onnx=={self.versions.ONNX} with pip"
                )
                logger.warning(msg)
            else:
                logger.info(
                    f"ONNX {self.versions.ONNX} installed in Python environment"
                )
        except SetupError as e:
            logger.warning(str(e))

    def check_tf_install(self):
        """Check Python environment for TensorFlow installation"""

        try:
            if not self.build_env.check_installed(
                "tensorflow", self.versions.TENSORFLOW
            ):
                msg = (
                    f"TensorFlow {self.versions.TENSORFLOW} not installed in Python environment. "
                    + f"Consider installing tensorflow=={self.versions.TENSORFLOW} with pip"
                )
                logger.warning(msg)
            else:
                logger.info(
                    f"TensorFlow {self.versions.TENSORFLOW} installed in Python environment"
                )
        except SetupError as e:
            logger.warning(str(e))



# BSD 2-Clause License
#
# Copyright (c) 2021, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import sys
from ..config import CONFIG
from ..utils.helpers import create_lockfile_name


def write_colocated_launch_script(file_name, db_log, colocated_settings):
    """Write the colocated launch script

    This file will be written into the cwd of the step that
    is created for this entity.

    :param file_name: name of the script to write
    :type file_name: str
    :param db_log: log file for the db
    :type db_log: str
    :param colocated_settings: db settings from entity run_settings
    :type colocated_settings: dict[str, Any]
    """

    colocated_cmd = _build_colocated_wrapper_cmd(**colocated_settings,
                                                 db_log=db_log)

    with open(file_name, "w") as f:
        f.write("#!/bin/bash\n")
        f.write("set -e\n\n")

        f.write("Cleanup () {\n")
        f.write("if ps -p $DBPID > /dev/null; then\n")
        f.write("\tkill -15 $DBPID\n")
        f.write("fi\n}\n\n")

        # run cleanup after all exitcodes
        f.write("trap Cleanup exit\n\n")

        # force entrypoint to write some debug information to the
        # STDOUT of the job
        if colocated_settings["debug"]:
            f.write("export SMARTSIM_LOG_LEVEL=debug\n")

        f.write(f"{colocated_cmd}\n")
        f.write(f"DBPID=$!\n\n")
        if colocated_settings["limit_app_cpus"]:
            cpus = colocated_settings["cpus"]
            f.write(
                f"taskset -c 0-$(nproc --ignore={str(cpus+1)}) $@\n\n"
            )
        else:
            f.write(f"$@\n\n")


def _build_colocated_wrapper_cmd(port=6780,
                                cpus=1,
                                interface="lo",
                                rai_args=None,
                                extra_db_args=None,
                                db_log=None,
                                **kwargs):
    """Build the command use to run a colocated db application

    :param port: db port, defaults to 6780
    :type port: int, optional
    :param cpus: db cpus, defaults to 1
    :type cpus: int, optional
    :param interface: network interface, defaults to "lo"
    :type interface: str, optional
    :param rai_args: redisai args, defaults to None
    :type rai_args: dict[str, str], optional
    :param extra_db_args: extra redis args, defaults to None
    :type extra_db_args: dict[str, str], optional
    :return: the command to run
    :rtype: str
    """

    # create unique lockfile name to avoid symlink vulnerability
    # this is the lockfile all the processes in the distributed
    # application will try to acquire. since we use a local tmp
    # directory on the compute node, only one process can acquire
    # the lock on the file.
    lockfile = create_lockfile_name()

    # create the command that will be used to launch the
    # database with the python entrypoint for starting
    # up the backgrounded db process
    cmd = [sys.executable,
           "-m",
           "smartsim._core.entrypoints.colocated",
           "+ifname",
           interface,
           "+lockfile",
           lockfile,
           "+db_cpus",
           str(cpus),
           "+command"
        ]

    # collect DB binaries and libraries from the config
    db_cmd = [
        CONFIG.redis_exe,
        CONFIG.redis_conf,
        "--loadmodule",
        CONFIG.redisai
    ]
    # add extra redisAI configurations
    for arg, value in rai_args.items():
        if value:
            # RAI wants arguments for inference in all capps
            # ex. THREADS_PER_QUEUE=1
            db_cmd.append(f"{arg.upper()} {str(value)}")

    # add port and log information
    db_cmd.extend([
        "--port",
        str(port),
        "--logfile",
        db_log # usually /dev/null
    ])
    for db_arg, value in extra_db_args.items():
        # replace "_" with "-" in the db_arg because we use kwargs
        # for the extra configurations and Python doesn't allow a hypon
        # in a variable name. All redis and KeyDB configuration options
        # use hyphens in their names.
        db_arg = db_arg.replace("_", "-")
        db_cmd.extend([
            f"--{db_arg}",
            value
        ])
    # run colocated db in the background
    db_cmd.append("&")

    cmd.extend(db_cmd)
    return " ".join(cmd)




import subprocess
from pathlib import Path

from smartsim._core._install.buildenv import SetupError
from smartsim._core._install.builder import BuildError
from smartsim._core.utils import colorize
from smartsim.log import get_logger

smart_logger_format = "[%(name)s] %(levelname)s %(message)s"
logger = get_logger("Smart", fmt=smart_logger_format)


def get_install_path():
    try:
        import smartsim as _
    except (ImportError, ModuleNotFoundError):
        raise SetupError("Could not import SmartSim") from None

    # find the path to the setup script
    package_path = Path(_.__path__[0]).resolve()
    if not package_path.is_dir():
        raise SetupError("Could not find SmartSim installation site")

    return package_path


def color_bool(trigger=True):
    _color = "green" if trigger else "red"
    return colorize(str(trigger), color=_color)


def pip_install(packages, end_point=None, verbose=False):
    """Install a pip package to be used in the SmartSim build
    Currently only Torch shared libraries are re-used for the build
    """
    if end_point:
        packages.append(f"-f {end_point}")
    packages = " ".join(packages)

    # form pip install command
    cmd = ["python", "-m", "pip", "install", packages]
    cmd = " ".join(cmd)

    if verbose:
        logger.info(f"Installing packages {packages}...")
    proc = subprocess.Popen(
        cmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE
    )
    _, err = proc.communicate()
    returncode = int(proc.returncode)
    if returncode != 0:
        error = f"{packages} installation failed with exitcode {returncode}\n"
        error += err.decode("utf-8")
        raise BuildError(error)
    if verbose:
        logger.info(f"{packages} installed successfully")



# BSD 2-Clause License
#
# Copyright (c) 2021-2022, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import time

from ....error import LauncherError
from ....log import get_logger
from ....settings import *
from ....status import STATUS_CANCELLED, STATUS_COMPLETED
from ..launcher import WLMLauncher
from ..step import AprunStep, LocalStep, MpirunStep, QsubBatchStep
from ..stepInfo import PBSStepInfo
from .pbsCommands import qdel, qstat
from .pbsParser import parse_qstat_jobid, parse_step_id_from_qstat

logger = get_logger(__name__)


class PBSLauncher(WLMLauncher):
    """This class encapsulates the functionality needed
    to launch jobs on systems that use PBS as a workload manager.

    All WLM launchers are capable of launching managed and unmanaged
    jobs. Managed jobs are queried through interaction with with WLM,
    in this case PBS. Unmanaged jobs are held in the TaskManager
    and are managed through references to their launching process ID
    i.e. a psutil.Popen object
    """

    # init in WLMLauncher, launcher.py

    # RunSettings types supported by this launcher
    supported_rs = {
        AprunSettings: AprunStep,
        QsubBatchSettings: QsubBatchStep,
        MpirunSettings: MpirunStep,
        RunSettings: LocalStep,
    }

    def run(self, step):
        """Run a job step through PBSPro

        :param step: a job step instance
        :type step: Step
        :raises LauncherError: if launch fails
        :return: job step id if job is managed
        :rtype: str
        """
        if not self.task_manager.actively_monitoring:
            self.task_manager.start()

        cmd_list = step.get_launch_cmd()
        step_id = None
        task_id = None
        if isinstance(step, QsubBatchStep):
            # wait for batch step to submit successfully
            rc, out, err = self.task_manager.start_and_wait(cmd_list, step.cwd)
            if rc != 0:
                raise LauncherError(f"Qsub batch submission failed\n {out}\n {err}")
            if out:
                step_id = out.strip()
                logger.debug(f"Gleaned batch job id: {step_id} for {step.name}")
        else:
            # aprun doesn't direct output for us.
            out, err = step.get_output_files()
            output = open(out, "w+")
            error = open(err, "w+")
            task_id = self.task_manager.start_task(
                cmd_list, step.cwd, out=output, err=error
            )

        # if batch submission did not successfully retrieve job ID
        if not step_id and step.managed:
            step_id = self._get_pbs_step_id(step)
        self.step_mapping.add(step.name, step_id, task_id, step.managed)

        return step_id

    def stop(self, step_name):
        """Stop/cancel a job step

        :param step_name: name of the job to stop
        :type step_name: str
        :return: update for job due to cancel
        :rtype: StepInfo
        """
        stepmap = self.step_mapping[step_name]
        if stepmap.managed:
            qdel_rc, _, err = qdel([str(stepmap.step_id)])
            if qdel_rc != 0:
                logger.warning(f"Unable to cancel job step {step_name}\n {err}")
            if stepmap.task_id:
                self.task_manager.remove_task(stepmap.task_id)
        else:
            self.task_manager.remove_task(stepmap.task_id)

        _, step_info = self.get_step_update([step_name])[0]
        step_info.status = STATUS_CANCELLED  # set status to cancelled instead of failed
        return step_info

    def _get_pbs_step_id(self, step, interval=2, trials=5):
        """Get the step_id of a step from qstat (rarely used)

        Parses qstat JSON output by looking for the step name
        TODO: change this to use ``qstat -a -u user``
        """
        time.sleep(interval)
        step_id = "unassigned"
        while trials > 0:
            output, _ = qstat(["-f", "-F", "json"])
            step_id = parse_step_id_from_qstat(output, step.name)
            if step_id:
                break
            else:
                time.sleep(interval)
                trials -= 1
        if not step_id:
            raise LauncherError("Could not find id of launched job step")
        return step_id

    def _get_managed_step_update(self, step_ids):
        """Get step updates for WLM managed jobs

        :param step_ids: list of job step ids
        :type step_ids: list[str]
        :return: list of updates for managed jobs
        :rtype: list[StepInfo]
        """
        updates = []

        qstat_out, _ = qstat(step_ids)
        stats = [parse_qstat_jobid(qstat_out, str(step_id)) for step_id in step_ids]
        # create PBSStepInfo objects to return

        for stat, _ in zip(stats, step_ids):
            info = PBSStepInfo(stat, None)
            # account for case where job history is not logged by PBS
            if info.status == STATUS_COMPLETED:
                info.returncode = 0

            updates.append(info)
        return updates

    def __str__(self):
        return "PBS"



# BSD 2-Clause License
#
# Copyright (c) 2021, Hewlett Packard Enterprise
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice, this
#    list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
#    this list of conditions and the following disclaimer in the documentation
#    and/or other materials provided with the distribution.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import os
import signal
import sys
import psutil
import argparse
import tempfile
import filelock
import socket

from typing import List
from pathlib import Path
from subprocess import PIPE, STDOUT

from smartsim._core.utils.network import current_ip
from smartsim.error import SSInternalError
from smartsim.log import get_logger
logger = get_logger(__name__)

DBPID = None

# kill is not catchable
SIGNALS = [
    signal.SIGINT,
    signal.SIGTERM,
    signal.SIGQUIT,
    signal.SIGABRT
    ]

def handle_signal(signo, frame):
    cleanup()


def main(network_interface: str, db_cpus: int, command: List[str]):
    global DBPID

    try:
        ip_address = current_ip(network_interface)
        lo_address = current_ip("lo")
    except ValueError as e:
        logger.warning(e)
        ip_address = None


    if lo_address == ip_address or not ip_address:
        cmd = command + [f"--bind {lo_address}"]
    else:
        # bind to both addresses if the user specified a network
        # address that exists and is not the loopback address
        cmd = command + [f"--bind {lo_address} {ip_address}"]

    # we generally want to catch all exceptions here as
    # if this process dies, the application will most likely fail
    try:
        p = psutil.Popen(cmd, stdout=PIPE, stderr=STDOUT)
        DBPID = p.pid

    except Exception as e:
        cleanup()
        logger.error(f"Failed to start database process: {str(e)}")
        raise SSInternalError("Co-located process failed to start") from e

    try:
        if sys.platform != "darwin":
            # Set CPU affinity to the last $db_cpus CPUs
            affinity = p.cpu_affinity()
            cpus_to_use = affinity[-db_cpus:]
            p.cpu_affinity(cpus_to_use)
        else:
            # psutil doesn't support pinning on MacOS
            cpus_to_use = "CPU pinning disabled on MacOS"

        logger.debug("\n\nCo-located database information\n" +  "\n".join((
            f"\tIP Address: {ip_address}",
            f"\t# of Database CPUs: {db_cpus}",
            f"\tAffinity: {cpus_to_use}",
            f"\tCommand: {' '.join(cmd)}\n\n"
        )))

        for line in iter(p.stdout.readline, b""):
            print(line.decode("utf-8").rstrip(), flush=True)

    except Exception as e:
        cleanup()
        logger.error(f"Co-located database process failed: {str(e)}")
        raise SSInternalError("Co-located entrypoint raised an error") from e


def cleanup():
    global DBPID
    global LOCK
    try:
        logger.debug("Cleaning up co-located database")
        # attempt to stop the database process
        db_proc = psutil.Process(DBPID)
        db_proc.terminate()

    except psutil.NoSuchProcess:
        logger.warning("Couldn't find database process to kill.")

    except OSError as e:
        logger.warning(
            f"Failed to clean up co-located database gracefully: {str(e)}"
        )
    finally:
        if LOCK.is_locked:
            LOCK.release()

        if os.path.exists(LOCK.lock_file):
            os.remove(LOCK.lock_file)


if __name__ == "__main__":
    try:
        parser = argparse.ArgumentParser(
            prefix_chars="+", description="SmartSim Process Launcher"
        )
        parser.add_argument("+ifname", type=str, help="Network Interface name", default="lo")
        parser.add_argument("+lockfile", type=str, help="Filename to create for single proc per host")
        parser.add_argument("+db_cpus", type=int, default=2, help="Number of CPUs to use for DB")
        parser.add_argument("+command", nargs="+", help="Command to run")
        args = parser.parse_args()

        tmp_lockfile = Path(tempfile.gettempdir()) / args.lockfile

        LOCK = filelock.FileLock(tmp_lockfile)
        LOCK.acquire(timeout=.1)
        logger.debug(f"Starting co-located database on host: {socket.gethostname()}")

        os.environ["PYTHONUNBUFFERED"] = "1"

        # make sure to register the cleanup before the start
        # the proecss so our signaller will be able to stop
        # the database process.
        for sig in SIGNALS:
            signal.signal(sig, handle_signal)

        main(args.ifname, args.db_cpus, args.command)

    # gracefully exit the processes in the distributed application that
    # we do not want to have start a colocated process. Only one process
    # per node should be running.
    except filelock.Timeout:
        exit(0)



from ..._core._install.buildenv import Version_, Versioner
from ...error import SmartSimError
from ...log import get_logger

logger = get_logger(__name__)

vers = Versioner()
TF_VERSION = vers.TENSORFLOW

try:
    import tensorflow as tf

    installed_tf = Version_(tf.__version__)
    assert installed_tf >= "2.4.0"

except ImportError:  # pragma: no cover
    raise ModuleNotFoundError(
        f"TensorFlow {TF_VERSION} is not installed. Please install it to use smartsim.tf"
    ) from None
except AssertionError:  # pragma: no cover
    raise SmartSimError(
        f"TensorFlow >= {TF_VERSION} is required for smartsim.tf, you have {tf.__version__}"
    ) from None


from .data import DynamicDataGenerator, StaticDataGenerator
from .utils import freeze_model



#!/usr/bin/env python

import argparse
import sys

from smartsim._core._cli.build import Build
from smartsim._core._cli.clean import Clean
from smartsim._core._cli.utils import get_install_path


def _usage():
    usage = [
        "smart <command> [<args>]\n",
        "Commands:",
        "\tbuild     Build SmartSim dependencies (Redis, RedisAI, ML runtimes)",
        "\tclean     Remove previous ML runtime installation",
        "\tclobber   Remove all previous dependency installations",
        "\nDeveloper:",
        "\tsite      Print the installation site of SmartSim",
        "\tdbcli     Print the path to the redis-cli binary" "\n\n",
    ]
    return "\n".join(usage)


class SmartCli:
    def __init__(self):
        parser = argparse.ArgumentParser(
            description="SmartSim command line interface", usage=_usage()
        )

        parser.add_argument("command", help="Subcommand to run")

        # smart
        if len(sys.argv) < 2:
            parser.print_help()
            exit(0)

        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            parser.print_help()
            exit(0)
        getattr(self, args.command)()

    def build(self):
        Build()
        exit(0)

    def clean(self):
        Clean()
        exit(0)

    def clobber(self):
        Clean(clean_all=True)
        exit(0)

    def site(self):
        print(get_install_path())
        exit(0)

    def dbcli(self):
        install_path = get_install_path()
        script_path = install_path.joinpath("_core/bin/redis-cli").resolve()
        if script_path.is_file():
            print(script_path)
            exit(0)
        else:
            print("Redis dependencies not found")
            exit(1)


def main():
    SmartCli()



from pathlib import Path

import tensorflow as tf
from tensorflow.python.framework.convert_to_constants import (
    convert_variables_to_constants_v2,
)


def freeze_model(model, output_dir, file_name):
    """Freeze a Keras or TensorFlow Graph

    to use a Keras or TensorFlow model in SmartSim, the model
    must be frozen and the inputs and outputs provided to the
    smartredis.client.set_model_from_file() method.

    This utiliy function provides everything users need to take
    a trained model and put it inside an ``orchestrator`` instance

    :param model: TensorFlow or Keras model
    :type model: tf.Module
    :param output_dir: output dir to save model file to
    :type output_dir: str
    :param file_name: name of model file to create
    :type file_name: str
    :return: path to model file, model input layer names, model output layer names
    :rtype: str, list[str], list[str]
    """
    # TODO figure out why layer names don't match up to
    # specified name in Model init.

    if not file_name.endswith(".pb"):
        file_name = file_name + ".pb"

    full_model = tf.function(lambda x: model(x))
    full_model = full_model.get_concrete_function(
        tf.TensorSpec(model.inputs[0].shape, model.inputs[0].dtype)
    )

    frozen_func = convert_variables_to_constants_v2(full_model)
    frozen_func.graph.as_graph_def()

    input_names = [x.name.split(":")[0] for x in frozen_func.inputs]
    output_names = [x.name.split(":")[0] for x in frozen_func.outputs]

    tf.io.write_graph(
        graph_or_graph_def=frozen_func.graph,
        logdir=output_dir,
        name=file_name,
        as_text=False,
    )
    model_file_path = str(Path(output_dir, file_name).resolve())
    return model_file_path, input_names, output_names
